
package java.lang;

import java.lang.ref.SoftReference;
import java.lang.reflect.*;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.WeakHashMap;
import java.security.AccessControlContext;
import java.security.ProtectionDomain;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.lang.reflect.Proxy;
import java.util.concurrent.ThreadPoolExecutor;
import com.ibm.tenant.ParameterizedRunnable;
import com.ibm.tenant.ReturnValue;
import com.ibm.tenant.SlotType;
import com.ibm.tenant.TenantContext;
import com.ibm.tenant.TenantException;
import com.ibm.tenant.TenantGlobals;
import com.ibm.tenant.internal.TenantAccessFactory;
import com.ibm.jvm.packed.PackedObject;
import com.ibm.jvm.packed.reflect.PackedArray;
import com.ibm.jvm.packed.reflect.PackedField;
import com.ibm.oti.vm.VM;

/*******************************************************************************
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 *     Copyright IBM Corp. 1998, 2013 All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

class J9VMInternals {

	private static final int j9Version = 0x05030270;

	private static final long j9Config = 0x7363617237306200L;	// 'scar70b\0'

	private static final int UNINITIALIZED = 0;
	private static final int INITIALIZED = 1;
	private static final int FAILED = 2;
	private static final int UNVERIFIED = 3;
	private static final int UNPREPARED = 4;
	private static final int INITIALIZED_TENANT = 5; // indicates this is a scoped class
	private static final int STATUS_MASK = 255;

	/* Constant filled in by the VM if multitenancy is enabled, equivalent to TenantGlobals.isTenantEnabled() */
	static final boolean MT_ENABLED;

	/* Constant filled in by the VM if tenant debug is enabled, equivalent to TenantGlobals.isDebugEnabled()  */
	private static final boolean MT_DEBUG_ENABLED;

	/* Constant filled in by the VM if tenant JCL support is enabled, equivalent to TenantGlobals.isMTJCLEnabled()  */
	private static final boolean MT_JCL_ENABLED;

	private static ParameterizedRunnable<Void, Object> tenantFinalizeRunnable;

	private static ParameterizedRunnable<Void, Class> initializeImplRunnable;

	/**
	 * @return A human readable version of the status constant.  Values in-lined
	 * to minimize dependencies while printing.
	 */
	@SuppressWarnings("nls")
	private static String decodeInitStatus(int status) {
		switch (status) {
		case UNINITIALIZED:
			return "UNINITIALIZED [0]";
		case INITIALIZED:
			return "INITIALIZED [1]";
		case FAILED:
			return "FAILED [2]";
		case UNVERIFIED:
			return "UNVERIFIED [3]";
		case UNPREPARED:
			return "UNPREPARED [4]";
		case INITIALIZED_TENANT:
			return "INITIALIZED_TENANT [5]";
		default:
			return VM.concatenate("OTHER [", Integer.toString(status), "]");
		}
	}

	// cannot create any instances in <clinit> in this special class
	private static Map exceptions;

	private static boolean initialized;

	static {
		/* Note this is never called - the VM marks this class as initialized immediately after loading.
		 * The initializer is here solely to trick the compiler into letting us have static final
		 * variables that are initialized by the VM.
		 */
		MT_ENABLED = false;
		MT_DEBUG_ENABLED = false;
		MT_JCL_ENABLED = false;
	}

	private static Object getInitializationLock(Class clazz)
	{
		return clazz.initializationLock;
	}

	private static void removeInitializationLock(Class clazz)
	{
		clazz.initializationLock = null;
	}

	/*
	 * Called by the vm after everything else is initialized.
	 */
	private static void completeInitialization() {
		initialized = true;
		exceptions = new WeakHashMap();

		ClassLoader.completeInitialization();
		Thread.currentThread().completeInitialization();
		if (MT_JCL_ENABLED){
			sendTenantCompleteInitialization();

			/* The following two ParameterizedRunnables must be allocated from the ROOT tenant. */
			tenantFinalizeRunnable = new ParameterizedRunnable<Void,Object>() {
				public Void run(Object runnableData) {
					try {
						/* temporary workaround until JAZZ103 46481 is resolved */
						if (runnableData instanceof ThreadPoolExecutor) {
							ThreadPoolExecutor tpe = (ThreadPoolExecutor)runnableData;
							if (tpe.isShutdown()) {
								return null; //do not run finalizer for this object
							}
						}
						runnableData.finalize();
					} catch (java.lang.Throwable e) {}
					return null;
				}
			};

			initializeImplRunnable = new ParameterizedRunnable<Void, Class>() {
				public Void run(Class clazz) {
					try {
						initializeImpl(clazz);
					} catch (java.lang.Throwable t) {
						throw new ExceptionInInitializerError(t); /* let the outer catch clause handle this throwable */
					}
					return null;
				}
			};
		}
	}

	private static native void sendClassPrepareEvent(Class clazz);

	/**
	 * Check whether -Xaggressive is option is on or not.
	 *
	 * @return true if -Xaggressive is on.
	 * 		   false if -Xaggressive is off.
	 */
	static native boolean isXaggressiveImpl();

	/**
	 * Verify the specified Class using the VM byte code verifier.
	 *
	 * @param clazz the Class to verify.
	 *
	 * @throws VerifyError if the Class cannot be verified
	 */
	static void verify(Class clazz) {

		if (MT_DEBUG_ENABLED && isTenantAware(clazz)) {
			VM.tenantDebug(
					"J9VMInternals.verify(",
					clazz.getName(),
					")");

			if (clazz.getName().equals("java.lang.IllegalMonitorStateException")) {
				dumpStack();
			}
		}

		Object initializationLock = getInitializationLock(clazz);
		if (initializationLock == null) {
			/* initializationLock == null means initialization successfully completed */
			return;
		}

		boolean isPendingInterrupt = false;
		try {
		while (true) {
			switch (getInitStatus(clazz)) {
				case INITIALIZED_TENANT:
				case INITIALIZED:
				case UNINITIALIZED:
				case FAILED:
				case UNPREPARED:
					return;

				case UNVERIFIED: {
					Class superclass;

					synchronized (initializationLock) {
						if (getInitStatus(clazz) != UNVERIFIED) break;
						setInitThread(clazz);
					}
					superclass = getSuperclass(clazz);
					boolean succeeded = false;
					try {
						// verify the superclass
						if (superclass != null)
							verify(superclass);
						// verify this class
						verifyImpl(clazz);
						succeeded = true;
					} finally {
						if (!succeeded) {
							setInitStatus(clazz, UNVERIFIED);
						}
					}
					synchronized (initializationLock) {
						if (getInitThread(clazz) && ((getInitStatus(clazz) & STATUS_MASK) == UNVERIFIED)) {
							setInitStatus(clazz, UNPREPARED);
						}
					}
					return;
				}

				default: // INPROGRESS
					synchronized (initializationLock) {
						int status = getInitStatus(clazz);

						if (((status & ~STATUS_MASK) == 0)) break;
						if ((status & STATUS_MASK) != UNVERIFIED) return;
						if (!getInitThread(clazz)) {
							try { initializationLock.wait(); } catch (InterruptedException e) {
								isPendingInterrupt = true;
							}
							break;
						}
					}

					boolean succeeded = false;
					try {
						verifyImpl(clazz);
						succeeded = true;
					} finally {
						if (!succeeded) {
							setInitStatus(clazz, UNVERIFIED);
						}
					}
					synchronized (initializationLock) {
						if (getInitThread(clazz) && ((getInitStatus(clazz) & STATUS_MASK) == UNVERIFIED)) {
							setInitStatus(clazz, UNPREPARED);
						}
					}
			}
		}
		} finally {
			if (isPendingInterrupt) {
				interruptCurrentThread();
			}
		}
	}

	// verifyImpl may throw a Throwable, however its undeclared so the Throwable
	// does not need to be wrapped in an Error or RuntimeException
	private static native void verifyImpl(Class clazz);

	/**
	 * Create a descriptive NoClassDefFoundError to throw if an attempt is made
	 * to initialize a Class which has previously failed initialization.
	 */
	private static NoClassDefFoundError initializationAlreadyFailed(Class clazz) {
		NoClassDefFoundError notFound = new NoClassDefFoundError(clazz.getName() + " (initialization failure)"); //$NON-NLS-1$
		// if exceptions is null, we're initializing and running single threaded
		if (exceptions != null) {
			synchronized(exceptions) {
				SoftReference weakReason = (SoftReference)exceptions.get(clazz);
				if (weakReason != null) {
					Throwable reason = (Throwable)weakReason.get();
					if (reason != null) {
						reason = copyThrowable(reason);
						notFound.initCause(reason);
					}
				}
			}
		}
		return notFound;
	}

	/**
	 * Record the cause (Throwable) of a failed initialization for a Class.
	 * If the Throwable is an Error, return that, otherwise, wrap the Throwable
	 * in an ExceptionInInitializerError.
	 */
	private static Error recordInitializationFailure(Class clazz, Throwable err) {
		if (initialized) {
			// if exceptions is null, we're initializing and running single threaded
			if (exceptions == null)
				exceptions = new WeakHashMap();
			synchronized(exceptions) {
				Throwable cause = err;
				if (err instanceof ExceptionInInitializerError) {
					cause = ((ExceptionInInitializerError)err).getException();
				}
				exceptions.put(clazz, new SoftReference(copyThrowable(cause)));
			}
		}
		if (err instanceof Error) {
			return (Error)err;
		}
		return new ExceptionInInitializerError(err);
	}

	/**
	 * Sent internally by the VM to initiatiate
	 * initialization of the receiver.  See chapter
	 * 2.17.5 of the JVM Specification (2nd ed)
	 *
	 * @throws		Throwable
	 */
	private static void initialize(Class clazz) throws Throwable {
		Object initializationLock = getInitializationLock(clazz);
		if (initializationLock == null) {
			/* initializationLock == null means initialization successfully completed */
			return;
		}

		boolean isPendingInterrupt = false;
		try {
		while (true) {
			switch (getInitStatus(clazz)) {
				case INITIALIZED:
					return;
				case INITIALIZED_TENANT:
					tenantInitialize(clazz);
					return;
				case UNVERIFIED:
					verify(clazz);
					break;
				case UNPREPARED:
					prepare(clazz);
					break;
				case FAILED:
					throw initializationAlreadyFailed(clazz);
				case UNINITIALIZED: {
					Class superclass;

					synchronized (initializationLock) {
						if (getInitStatus(clazz) != UNINITIALIZED) break;
						setInitThread(clazz);

						/* Before dropping the lock ensure that tenant slots have been allocated as
						 * superclass initialization may require running code on this tenant which
						 * demands that static fields have been set to default values.*/
						if (MT_JCL_ENABLED && isTenantAware(clazz)) {
							prepareTenant(clazz);
						}
					}

					// initialize the superclass
					superclass = getSuperclass(clazz);
					if (superclass != null) {
						try {
							initialize(superclass);
						} catch (Error err) {
							/* Record the initialization failure before changing the class init
							 * status.  This means that any threads woken up by the change in
							 * status will be able to fetch the more descriptive error.
							 */
							Error newError = recordInitializationFailure(clazz, err);
							setInitStatus(clazz, FAILED);
							throw newError;
						}
					}

					if (VM.PACKED_SUPPORT_ENABLED) {
						if (PackedObject.class.isAssignableFrom(clazz)) {
							if (PackedObject.isPackedArray(clazz)) {
								try {
									initialize(PackedArray.getComponentType(clazz));
								} catch (Error err) {
									/* Record the initialization failure before changing the class init
									 * status.  This means that any threads woken up by the change in
									 * status will be able to fetch the more descriptive error.
									 */
									Error newError = recordInitializationFailure(clazz, err);
									setInitStatus(clazz, FAILED);
									throw newError;
								}
							} else {
								PackedField[] packedFields = PackedField.getDeclaredFields(clazz);
								for (PackedField pf : packedFields) {
									if (pf.isNestedPacked()) {
										try {
											initialize(pf.getType());
										} catch (Error err) {
											/* Record the initialization failure before changing the class init
											 * status.  This means that any threads woken up by the change in
											 * status will be able to fetch the more descriptive error.
											 */
											Error newError = recordInitializationFailure(clazz, err);
											setInitStatus(clazz, FAILED);
											throw newError;
										}
									}
								}
							}
						}
					}

					// initialize this class
					try {
						boolean ranInitializeImpl = false;

						/* When we are initializing the statics of the class
						 * we want to be in the correct memory space for the class loader.
						 * If the class loader does not have a memory space, then we want
						 * to initialize it in the base memory space.  In the situation where
						 * we are not in multi-memory space mode, we will get null back when
						 * we request the memory space for the class loader, and we won't need
						 * to switch memory spaces.  If we have the system class loader then we
						 * want to initialize in the base memory space.  Their are problems treating
						 * the system class loader the same as others because not everything required
						 * by MemorySpace.getMemorySpaceForClassLoader has been initialized the first
						 * time the method is called.
						 */
						/**
						 * If clazz is not isolated, it's <clinit> should be run in the same
						 * tenant which owns the memory that the clazz was allocated in.
						 * An example of where this is necessary is when a non-isolated
						 * "filter" class is initialized by a regular tenant.
						 */
						if (MT_JCL_ENABLED && TenantGlobals.isRootTenantInitialized()) {
							if (!isTenantAware(clazz) ) {
								TenantContext clazzTenant = TenantContext.getTenantForObject(clazz);
								if (null != clazzTenant) {
									TenantContext targetTenant = TenantGlobals.getTenantContextIfDifferent(clazzTenant); /* check if we're already running in the right tenant */
									if (null != targetTenant) {
										/* we're not running in the right tenant, switch into it and call <clinit> */
										TenantAccessFactory.getAccess().primitiveRun(targetTenant, initializeImplRunnable, clazz);
										ranInitializeImpl = true;
									}
								}
							}
						}

						if (!ranInitializeImpl) {
							initializeImpl(clazz);
						}
					} catch (Throwable t) {
						/* Record the initialization failure before changing the class init
						 * status.  This means that any threads woken up by the change in
						 * status will be able to fetch the more descriptive error.
						 */
						Error newError = recordInitializationFailure(clazz, t);
						setInitStatus(clazz, FAILED);
						throw newError;
					}
					if ( MT_JCL_ENABLED && isTenantAware(clazz) ) {
						setInitStatus(clazz, INITIALIZED_TENANT);
						break;
					} else {
						setInitStatus(clazz, INITIALIZED);
						removeInitializationLock(clazz);
						return;
					}
				}

				default: // INPROGRESS
					synchronized (initializationLock) {
						int status = getInitStatus(clazz);

						if ((status & ~STATUS_MASK) == 0) break;
						if ((status & STATUS_MASK) == UNINITIALIZED) {
							if (getInitThread(clazz)) {
								if ( MT_JCL_ENABLED && isTenantAware(clazz) ) {
									/* Set this status to indicate that the current class is ready to be initialized for other tenants.
									/* The case happened in such a situation:
									/* the <clinit> of class Foo of root tenant has triggered a call to the <clinit> of class Bar of tenant A,
									 * and inside the <clinit> of class Bar, a new instance of class Foo is created.
									 * we need set the init status of class Foo as INITIALIZED_TENANT here
									 * which prepares class Foo to be ready to be initialized for Tenant A.
									 */
									setInitStatus(clazz, INITIALIZED_TENANT);
									break;
								} else {
									return;
								}
							}
							try { initializationLock.wait(); } catch (InterruptedException e) {
								isPendingInterrupt = true;
								}
							break;
						}
					}
					verify(clazz);
			}
		}
		} finally {
			if (isPendingInterrupt) {
				interruptCurrentThread();
			}

		}
	}

	/**
	 * Debug helper to dump the current call stack.
	 */
	private static void dumpStack() {
		Exception e = new Exception();
		e.fillInStackTrace();

		StackTraceElement[] frames = e.getInternalStackTrace();
		String[] printedFrames = new String[frames.length+1];
		int i=0;

		printedFrames[i++] = "J9VMInternals.dumpStack(): \n";

		for (StackTraceElement frame : e.getInternalStackTrace()) {
			String printedFrame =
				VM.concatenate(
					"    ",
					frame.getClassName(),
					".",
					frame.getMethodName(),
					"(",
					(frame.getFileName() == null ? "<no file>" : frame.getFileName()),
					":",
					(frame.getLineNumber() < 0 ? "<no line>" : Integer.toString(frame.getLineNumber())),
					")\n");
			printedFrames[i++] = printedFrame;
		}
		VM.tenantDebug(printedFrames);
	}

	/**
	 * Lifecycle hook called during tenant creation to initialize the class
	 * library.
	 * @param tenant
	 * @throws TenantException
	 */
	static boolean initializeSystem(TenantContext tenant) throws TenantException {
		// Creating a HashMap is the usual way to trigger Oracle class library initialization
		HashMap m = new HashMap();
		System.afterClinitInitialization();
		System.completeInitialization();
		return com.ibm.tenant.internal.TenantAccessFactory.getAccess().completeTenantInitialization(tenant);
	}

	/**
	 * Initialize the receiver for the current tenant.
	 *
	 * @param clazz	the receiver
	 * @throws java.lang.Throwable
	 */
	@SuppressWarnings("nls")
	private static void tenantInitialize(java.lang.Class clazz) throws java.lang.Throwable {

		try {
			if (MT_DEBUG_ENABLED && isTenantAware(clazz)) {
				VM.tenantDebug(
						"J9VMInternals.tenantInitialize(",
						clazz.getName(),
						")");
			}

			if ( !isTenantAware(clazz) ) {
				// Class is not tenant aware, but it's superclass might be
				Class superclass;
				superclass = getSuperclass(clazz);
				if ( superclass != null && superclass != Object.class) {
					tenantInitialize(superclass);
				}
				return;
			}

			Object initializationLock = getInitializationLock(clazz);

			while (true) {
				int status = getTenantInitStatus(clazz);
				switch ( status ) {

				case INITIALIZED:
					if (MT_DEBUG_ENABLED) {
						VM.tenantDebug(
							"J9VMInternals.tenantInitialize(",
								clazz.getName(),
								") -> already INITIALIZED.");
						dumpStack();
					}
					return;

				case FAILED:
					NoClassDefFoundError notFound = new NoClassDefFoundError(clazz.getName() + " (tenant initialization failure)"); //$NON-NLS-1$
					throw notFound;

				case UNINITIALIZED:
					synchronized (initializationLock) {

						// Get the init status under lock
						int statusUnderLock = getTenantInitStatus(clazz);
						if ( statusUnderLock != UNINITIALIZED) {
							if (MT_DEBUG_ENABLED) {
								boolean initializingOnThisThread = getTenantInitThread(clazz);
								VM.tenantDebug(
										"J9VMInternals.tenantInitialize(",
										clazz.getName(),
										") -> breaking with initStatus=",
										decodeInitStatus(statusUnderLock),
										", thread=",
										(initializingOnThisThread ? "true" : "false") );
								dumpStack();
							}

							break;	// Fall through to default case
						}

						// We own the initialization lock at this point, ensure that we have tenant data
						int[] slotCounts = new int[SlotType.TOTAL_SLOT_TYPES];
						getTenantSlotCounts(clazz, slotCounts);
						int[][] slots = new int[SlotType.TOTAL_SLOT_TYPES][];
						slots[SlotType.B32] = new int[slotCounts[SlotType.B32]];
						slots[SlotType.B64] = new int[slotCounts[SlotType.B64]];
						slots[SlotType.REF] = new int[slotCounts[SlotType.REF]];
						slots[SlotType.B64_INITSTATUS] = new int[slotCounts[SlotType.B64_INITSTATUS]];
						getTenantSlots(clazz, slots);

						// make sure the slots allocated and initialized for the current tenant
						com.ibm.tenant.internal.TenantAccessFactory.getAccess().ensureAllocationAndInitializeDefaultsForTenant(slots);
						if (MT_DEBUG_ENABLED) {
							VM.tenantDebug(
									"J9VMInternals.tenantInitialize(",
									clazz.getName(),
									") -> initializing on this thread");
						}
						setTenantInitThread(clazz);
					}

					// initialize the superclass
					Class superclass = getSuperclass(clazz);

					if ( superclass != null && superclass != Object.class) { //TODO work out a better way to not loop forever on Object
						try {
							tenantInitialize(superclass);
						} catch (Error err) {
							setTenantInitStatus(clazz, FAILED);
							throw err;
						}
					}

					try {
						tenantInitializeImpl(clazz);
						setTenantInitStatus(clazz, INITIALIZED);
					} catch (Error err) {
						setTenantInitStatus(clazz, FAILED);
						throw err;
					} catch (Throwable t) {
						setTenantInitStatus(clazz, FAILED);
						throw new ExceptionInInitializerError(t);
					}
					return;

				default:
					synchronized (initializationLock) {
						int tenantStatus = getTenantInitStatus(clazz);

						// No thread currently marking this Tenant's clazz as being in progress, init likely complete, break out and check
						if ((tenantStatus & ~STATUS_MASK) == 0) {
							if (MT_DEBUG_ENABLED) {
								boolean initializingOnThisThread = getTenantInitThread(clazz);
								VM.tenantDebug(
										"J9VMInternals.tenantInitialize(",
										clazz.getName(),
										") -> breaking with initStatus=",
										decodeInitStatus(tenantStatus),
										", initializingOnThisThread=",
										(initializingOnThisThread ? "true" : "false") );
								dumpStack();
							}
							break;
						}

						// A thread is tenant initializing this clazz
						if ((tenantStatus & STATUS_MASK) == UNINITIALIZED) {

							// class is still being initialized
							if (getTenantInitThread(clazz)) {
								if (MT_DEBUG_ENABLED) {
									boolean initializingOnThisThread = getTenantInitThread(clazz);
									VM.tenantDebug(
											"J9VMInternals.tenantInitialize(",
											clazz.getName(),
											") -> recursive initialization in progress initStatus=",
											decodeInitStatus(tenantStatus),
											", initializingOnThisThread=",
											(initializingOnThisThread ? "true" : "false") );
								}
								return;  // apparently by this thread return so it can complete.
							}

							// another thread is doing init, wait until complete or interrupt and check the state again.
							try { initializationLock.wait(); } catch (InterruptedException e) {}
							continue;
						}
					}
				}
			}
		} finally {
			if (MT_DEBUG_ENABLED && isTenantAware(clazz)) {
				int initStatus = getTenantInitStatus(clazz);
				VM.tenantDebug(
						"< J9VMInternals.tenantInitialize(",
						clazz.getName(),
						") -> status=",
						decodeInitStatus(initStatus));
			}
		}
	}

	/**
	 * Prepare the receiver for tenant operation.
	 * Allocate slot indices for isolated fields of the receiver.
	 * @param clazz
	 */
	private static void prepareTenant(java.lang.Class clazz) {

		if (MT_DEBUG_ENABLED) {
			VM.tenantDebug(
					"J9VMInternals.prepareTenant(",
					clazz.getName(),
					")");
		}

		final int[] slotCounts = new int[SlotType.TOTAL_SLOT_TYPES];
		getTenantSlotCounts(clazz, slotCounts);

		try {
			ReturnValue<int[][]> retVal = com.ibm.tenant.internal.TenantAccessFactory.getAccess().allocateSlots(slotCounts, clazz);
			associateTenantSlots(retVal.getResult(), clazz);
		} catch (TenantException e) {
			throw new ExceptionInInitializerError(e);
		}
	}

	/**
	 * Run the receiver's tclinit() function, initialize isolated fields for current tenant.
	 *
	 * @param clazz the receiver
	 * @throws Throwable
	 */
	private static native void tenantInitializeImpl(Class clazz) throws Throwable;

	/**
	 * Check whether the receiver is tenant aware. (Contains isolated fields)
	 *
	 * @param clazz   the receiver
	 * @return true if the receiver is tenant aware.
	 * 		   false if the receiver is just a normal class.
	 */
	static native boolean isTenantAware(Class clazz);

	/**
	 * Get the count of isolated fields(three types) of the receiver.
	 * @param clazz  the receiver
	 * @param an array of counts int[]{32-bit field count, 64-bit field count, object-ref field count}.
	 */
	private static native void getTenantSlotCounts(java.lang.Class clazz, int[] slotCounts);

	/**
	 * Assign the allocated slot indices to fields of the receiver.
	 *
	 * @param slots	 slot indices allocated by TenantDataRepo
	 * @param clazz	 the receiver
	 */
	private static native void associateTenantSlots(int[][] slots, java.lang.Class clazz);

	/**
	 * Get slot indices for isolated fields of the receiver.
	 *
	 * @param clazz  the receiver
	 * @param a double dimension array contains indices of three types of fields.
	 * 		   int[]{int[] 32-bit fields indices, int[] 64-bit fields indices, int[] object-ref fields indices}.
	 */
	private static native void getTenantSlots(java.lang.Class clazz, int[][] slots);

	/**
	 * Call "static void TenantContext.completeInitialization"
	 *
	 * We need to drop to JNI to make this call as completeInitialization has default visibility and is in a different package
	 */
	private static native void sendTenantCompleteInitialization();

	private static native Throwable newInstance(Class exceptionClass, Class constructorClass);

	private static Throwable cloneThrowable(final Throwable throwable, final HashMap hashMapThrowable) {
		return (Throwable)AccessController.doPrivileged(new PrivilegedAction() {
			public Object run() {
				Throwable clone;
				try {
					Class cls = throwable.getClass();
					clone = newInstance(cls, Object.class);
					while (cls != null) {
						Field[] fields = cls.getDeclaredFields();
						for (int i=0; i<fields.length; i++) {
							if (!Modifier.isStatic(fields[i].getModifiers()) &&
									!(cls == Throwable.class && fields[i].getName().equals("walkback")))
							{
								fields[i].setAccessible(true);
								Object value;
								if (cls == Throwable.class && fields[i].getName().equals("cause")) {
									value = clone;
								} else {
									value = fields[i].get(throwable);
									//	Only copy throwable fields whose stacktrace might be kept within Map exceptions
									//	The throwable stored within Map exceptions as WeakReference could be retrieved (before being GCed) later
									if (value instanceof Throwable) {
										value = copyThrowable((Throwable)value, hashMapThrowable);
									}
								}
								fields[i].set(clone, value);
							}
						}
						cls = getSuperclass(cls);
					}
				} catch (Throwable e) {
					clone = new Throwable("Error cloning Throwable (" + e + "). The original exception was: " + throwable.toString());
				}
				return clone;
			}
		});
	}

	/**
	 * Entry method to copy the specified Throwable, invoke copyThrowable(Throwable, HashMap)
	 * to check loop such that we don't go infinite.
	 *
	 * @param throwable the Throwable to copy
	 *
	 * @return a copy of the Throwable
	 */
	private static Throwable copyThrowable(Throwable throwable) {
		HashMap hashMapThrowable = new HashMap();
		return copyThrowable(throwable, hashMapThrowable);
	}

	/**
	 * Copy the specified Throwable, wrapping the stack trace for each
	 * Throwable. Check for loops so we don't go infinite.
	 *
	 * @param throwable the Throwable to copy
	 * @param hashMapThrowable the Throwables already cloned
	 *
	 * @return a copy of the Throwable or itself if it has been cloned already
	 */
	private static Throwable copyThrowable(Throwable throwable, HashMap hashMapThrowable) {
		if (hashMapThrowable.get(throwable) != null) {
			//	stop recursive call here when the throwable has been cloned
			return	throwable;
		}
		hashMapThrowable.put(throwable, throwable);
		Throwable root = cloneThrowable(throwable, hashMapThrowable);
		root.setStackTrace(throwable.getStackTrace());
		Throwable parent = root;
		Throwable cause = throwable.getCause();
		//	looking for causes recursively which will be part of stacktrace	stored into Map exceptions
		while (cause != null && hashMapThrowable.get(cause) == null) {
			hashMapThrowable.put(cause, cause);
			Throwable child = cloneThrowable(cause, hashMapThrowable);
			child.setStackTrace(cause.getStackTrace());
			parent.setCause(child);
			parent = child;
			cause = cause.getCause();
		}
		return root;
	}

	/**
	 * Used to indicate the end of class initialization.
	 * Sets the initialization status and notifies any
	 * threads which may have been waiting for
	 * initialization to complete
	 *
	 * @param		status
	 *					INITIALIZED (1)
	 *					FAILED (2)
	 *
	 */
	private static void setInitStatus(Class clazz, int status) {

		if (MT_DEBUG_ENABLED && isTenantAware(clazz)) {
			int oldStatus = getInitStatus(clazz);
			VM.tenantDebug(
					"J9VMInternals.setInitStatus(",
					clazz.getName(),
					", old=",
					decodeInitStatus(oldStatus),
					", new=",
					decodeInitStatus(status),
					")");
		}

		Object initializationLock = getInitializationLock(clazz);
		/* initializationLock == null means initialization successfully completed.
		 * Trying to change the init status after this is an error, so no need for
		 * a null check here.
		 *
		 */

		synchronized(initializationLock) {
			setInitStatusImpl(clazz, status);
			initializationLock.notifyAll();
		}
	}

	/**
	 * Answers the receiver's initialization status
	 *
	 * @return		status
	 *					UNINITIALIZED (0)
	 *					INITIALIZED (1)
	 *					FAILED (2)
	 *					INPROGRESS (0xnnnnnnn[048C])
	 *
	 */
	private static native int getInitStatus(Class clazz);

	/**
	 * Set the receiver's initialization status
	 *
	 * @param		status
	 *					INITIALIZED (1)
	 *					FAILED (2)
	 *
	 */
	private static native void setInitStatusImpl(Class clazz, int status);

	/**
	 * Run the receiver's <clinit> method and initialize
	 * any static variables
	 *
	 * @throws		Throwable Any exception may be thrown
	 */
	private static native void initializeImpl(Class clazz) throws Throwable;

	/**
	 * Answers true if the current thread is currently
	 * initializing this class
	 *
	 * @return		true if the current thread is initializing the receiver
	 *
	 */
	private static native boolean getInitThread(Class clazz);

	/**
	 * Set the receiver's initialize status to be 'in
	 * progress' and save the current thread as the
	 * initializing thread.
	 */
	private static native void setInitThread(Class clazz);

	/**
	 * Used to indicate the end of tenant class initialization.
	 * Sets the initialization status and notifies any
	 * threads which may have been waiting for
	 * initialization to complete
	 *
	 * @param		status
	 *					INITIALIZED (1)
	 *					FAILED (2)
	 *
	 */
	private static void setTenantInitStatus(Class clazz, int status) {
		if (MT_DEBUG_ENABLED && isTenantAware(clazz)) {
			int oldStatus = getTenantInitStatus(clazz);
			VM.tenantDebug(
					"J9VMInternals.setTenantInitStatus(",
					clazz.getName(),
					", old=",
					decodeInitStatus(oldStatus),
					", new=",
					decodeInitStatus(status),
					")");
		}

		Object initializationLock = getInitializationLock(clazz);
		synchronized(initializationLock) {
			setTenantInitStatusImpl(clazz, status);
			initializationLock.notifyAll();
		}
	}

	/**
	 * Answers the receiver's tenant initialization status
	 *
	 * @return		status
	 *					UNINITIALIZED (0)
	 *					INITIALIZED (1)
	 *					FAILED (2)
	 *					INPROGRESS (0xnnnnnnn[048C])
	 *
	 */
	private static native int getTenantInitStatus(Class clazz);

	/**
	 * Set the receiver's tenant initialization status
	 *
	 * @param		status
	 *					INITIALIZED (1)
	 *					FAILED (2)
	 *
	 */
	private static native void setTenantInitStatusImpl(Class clazz, int status);

	/**
	 * Answers true if the current thread is currently
	 * initializing this tenant's class
	 *
	 * @return		true if the current thread is initializing the receiver
	 *
	 */
	private static native boolean getTenantInitThread(Class clazz);

	/**
	 * Set the receiver's tenant initialize status to be 'in
	 * progress' and save the current thread as the
	 * initializing thread.
	 */
	private static native void setTenantInitThread(Class clazz);

	/**
	 * Private method to be called by the VM after a Threads dies and throws ThreadDeath
	 * It has to <code>notifyAll()</code> so that <code>join</code> can work properly.
	 * However, it has to be done when the Thread is "thought of" as being dead by other
	 * observer Threads (<code>isAlive()</code> has to return false for the Thread
	 * running this method or <code>join</code> may never return)
	 *
	 * @author		OTI
	 * @version		initial
	 */
	private static void threadCleanup(Thread thread) {
		// don't synchronize the remove! Otherwise deadlock may occur
		try {
			// Leave the ThreadGroup. This is why remove can't be private
			thread.threadGroup.remove(thread);
		}
		finally {
			thread.cleanup();

			synchronized(thread) {
				thread.notifyAll();
			}
		}
	}

	private static void checkPackageAccess(final Class clazz, ProtectionDomain pd) {
		final SecurityManager sm = System.getSecurityManager();
		if (sm != null) {
			AccessController.doPrivileged(new PrivilegedAction() {
				public Object run() {
					sm.checkPackageAccess(clazz.getPackageName());
					if (Proxy.isProxyClass(clazz)) {
						ClassLoader	cl = getClassLoader(clazz);
						sun.reflect.misc.ReflectUtil.checkProxyPackageAccess(cl, clazz.getInterfaces());
					}
					return null;
				}
			}, new AccessControlContext(new ProtectionDomain[]{pd}));
		}
	}

	private static void runFinalize(Object obj) {
		if (MT_JCL_ENABLED) {
			TenantContext tenant = TenantContext.getTenantForObject(obj);
			if (null == tenant) {
				try {
					obj.finalize();
				} catch(Throwable e) {}
			} else {
				try {
					TenantAccessFactory.getAccess().primitiveRun(tenant, tenantFinalizeRunnable, obj);
				} catch(Throwable e) {}
			}
		} else {
			try {
				obj.finalize();
			} catch(Throwable e) {}
		}
	}

	static native StackTraceElement[] getStackTrace(Throwable throwable, boolean pruneConstructors);

	private static native void prepareClassImpl(Class clazz);

	/**
	 * Prepare the specified Class. Fill in initial field values, and send
	 * the class prepare event.
	 *
	 * @param clazz the Class to prepare
	 */
	static void prepare(Class clazz) {
		if (MT_ENABLED) {
			prepareClass(clazz);
			return;
		}
		Object initializationLock = getInitializationLock(clazz);
		if (initializationLock == null) {
			/* initializationLock == null means initialization successfully completed */
			return;
		}
		prepareClassImpl(clazz);
	}

	/**
	 * Prepare the specified Class. Fill in initial field values, and send
	 * the class prepare event.
	 *
	 * @param clazz the Class to prepare
	 */
	private static void prepareClass(Class clazz) {

		if (MT_DEBUG_ENABLED && isTenantAware(clazz)) {
			VM.tenantDebug(
					"J9VMInternals.prepare(",
					clazz.getName(),
					")");
		}

		Object initializationLock = getInitializationLock(clazz);
		if (initializationLock == null) {
			/* initializationLock == null means initialization successfully completed */
			return;
		}

		while (true) {
			switch (getInitStatus(clazz)) {
				case INITIALIZED_TENANT:
				case INITIALIZED:
				case UNINITIALIZED:
				case FAILED:
					return;

				case UNVERIFIED:
					verify(clazz);
					break;

				case UNPREPARED: {
					Class superclass;

					superclass = getSuperclass(clazz);
					// prepare the superclass and direct superinterfaces
					if (superclass != null)
						prepare(superclass);
					Class interfaces[] = getInterfaces(clazz);
					for (int i = 0; i < interfaces.length; ++i) {
						prepare(interfaces[i]);
					}

					synchronized (initializationLock) {
						if (getInitStatus(clazz) != UNPREPARED) break;
						setInitStatus(clazz, UNINITIALIZED);
					}
					sendClassPrepareEvent(clazz);
					return;
				}

				default: // INPROGRESS
					// This cannot happen
					return;
			}
		}
	}

	/**
	 * Determines the superclass of specified <code>clazz</code>.
	 * @param clazz The class to introspect (must not be null).
	 * @return The superclass, or null for primitive types and interfaces.
	 */
	static native Class getSuperclass(Class clazz);

	/**
	 * Determines the interfaces implemented by <code>clazz</code>.
	 * @param clazz The class to introspect (must not be null).
	 * @return An array of all interfaces supported by <code>clazz</code>.
	 */
	static native Class[] getInterfaces(Class clazz);

	/**
	 * Determines the ClassLoader for <code>clazz</code>.
	 * @param clazz The class to introspect (must not be null).
	 * @return The ClassLoader which loaded <code>clazz</code>.
	 */
	static native ClassLoader getClassLoader(Class clazz);

	/**
	 * Answers a new instance of the class represented by the
	 * <code>clazz</code>, created by invoking the default (i.e. zero-argument)
	 * constructor. If there is no such constructor, or if the
	 * creation fails (either because of a lack of available memory or
	 * because an exception is thrown by the constructor), an
	 * InstantiationException is thrown. If the default constructor
	 * exists, but is not accessible from the context where this
	 * message is sent, an IllegalAccessException is thrown.
	 *
	 * @param clazz The class to create an instance of.
	 * @return		a new instance of the class represented by the receiver.
	 * @throws		IllegalAccessException if the constructor is not visible to the sender.
	 * @throws		InstantiationException if the instance could not be created.
	 */
	native static Object newInstanceImpl(Class clazz)
		throws IllegalAccessException, InstantiationException;

	/**
	 * Answers an integer hash code for the parameter.
	 * The hash code returned is the same one that would
	 * be returned by java.lang.Object.hashCode(), whether
	 * or not the object's class has overridden hashCode().
	 * Calling with null will cause a crash, so the caller
	 * must check for null.
	 *
	 * @param		anObject	the object
	 * @return		the hash code for the object
	 *
	 * @see			java.lang.Object#hashCode
	 */
	static native int identityHashCode(Object anObject);

	/**
	 * Primitive implementation of Object.clone().
	 * Calling with null will cause a crash, so the caller
	 * must check for null.
	 *
	 * @param		anObject	the object
	 * @return		a shallow copy of anObject
	 *
	 * @see			java.lang.Object#clone
	 */
	static native Object primitiveClone(Object anObject) throws CloneNotSupportedException;

	/**
	 * A class used for synchronizing Class initialization.
	 *
	 * The instance field is currently only for debugging purposes.
	 */
	static class ClassInitializationLock {
		Class theClass;
	}

	/**
	 * Native used to dump a string to the system console for debugging.
	 *
	 * @param 		str String
	 *					the String to display
	 */
	public static native void dumpString(String str);

	/**
	 * This implementation is copied from @ref System.arraycopy().
	 *
	 * This method is called from bootstrap code to avoid loading the
	 * per-tenant isolated System class in a multi-tenant VM before
	 * the class isolation mechanisms are completely initialized
	 */
	static void arraycopy(char[] A1, int offset1, char[] A2, int offset2, int length) {
		if (MT_DEBUG_ENABLED) {
			if (A1 == null || A2 == null) throw new NullPointerException();
			if (offset1 >= 0 && offset2 >= 0 && length >= 0 &&
					length <= A1.length - offset1 && length <= A2.length - offset2)
			{
				// Check if this is a forward or backwards arraycopy
				if (A1 != A2 || offset1 > offset2 || offset1+length <= offset2) {
					for (int i = 0; i < length; ++i) {
						A2[offset2+i] = A1[offset1+i];
					}
				} else {
					for (int i = length-1; i >= 0; --i) {
						A2[offset2+i] = A1[offset1+i];
					}
				}
			} else {
				throw new ArrayIndexOutOfBoundsException();
			}
		} else {
			System.arraycopy(A1, offset1, A2, offset2, length);
		}
	}

	/**
	 * Called by the VM class initialization code to repost an interrupt
	 * consumed by waiting on the initialization lock.
	 */
	private static void interruptCurrentThread() {
		Thread.currentThread().interrupt();
	}

}
